from ...support import interfaces

class BruteForceAndIgnorance(interfaces.AbstractGeneralizedParser):
	"""
	There is an old adage in software development: when in doubt, use brute force. Accordingly,
	the first implementation will be simple, easy to code, slow, and vulnerable to exponential
	behavior. However, it will also provide a fine basis of comparison for other approaches.
	
	This is the "inverse tree" approach: each node is a state, a predecessor, and a semantic
	value for the reaching-symbol -- except for the sink, which is special. The top-of-stack
	is just a list of currently-viable sub-stacks. This is usually sufficient for cases that
	are actually unambiguous but just not quite LR(1)-deterministic.
	
	Big problems with this approach, in no particular order, include:
		1. It fails to coalesce parse stacks that reduce to a common state.
		2. It eagerly calls reductions, even for dead-end parse attempts.
	
	So you can read what's going on:
		self.__tos contains nodes representing currently active parses.
		self.__next gets filled with the results of SHIFT actions.
		Each node is a tuple of (from_state_id, prior_node, semantic_value), accessed
			by field numbers given as the constants below.
	"""
	
	NODE_STATE = 0
	NODE_PRIOR = 1
	NODE_SEMANTIC = 2
	
	def reset(self, initial_state):
		self.__tos = [(initial_state, None, None)]
	
	def consume(self, terminal, semantic):
		self.__consume(self._table.get_translation(terminal), semantic)
		if not self.__tos: raise interfaces.GeneralizedParseError("Parser died midway at something ungrammatical.")

	def __consume(self, terminal_id, semantic):
		self.__next = []
		while self.__tos:
			top = self.__tos.pop()
			self.__act(self._table.get_action(top[self.NODE_STATE], terminal_id), top, semantic)
		self.__tos = self.__next

	def finish(self) -> list:
		"""
		Call this after the last token to wrap up and
		:return: a valid semantic value for the parse.
		"""
		self.__consume(0, None)
		if self.__tos: return [top[self.NODE_PRIOR][self.NODE_SEMANTIC] for top in self.__tos]
		else: raise interfaces.GeneralizedParseError("Parser recognized a viable prefix, but not a complete sentence.")
	
	def __act(self, action, top, semantic):
		""" There are four kinds of action: die, shift, reduce-and-shift, or split into parallel alternatives. """
		if action == 0: return # This branch of the stack dies.
		elif action < 0: self.__tos.append(self.__reduction(-1 - action, top))
		elif action < self._nr_states: self.__shift(action, top, semantic)
		else:
			for alternative in self._table.get_split(action - self._nr_states):
				self.__act(alternative, top, semantic)
	
	def __shift(self, state_id, top, semantic):
		shift = state_id, top, semantic
		while True:
			action = self._table.interactive_step(shift[self.NODE_STATE])
			if action < 0: shift = self.__reduction(-1 - action, shift)
			else: break
		self.__next.append(shift)
	
	def __reduction(self, rule_id, top):
		nonterminal_id, length, cid, view = self._table.get_rule(rule_id)
		if cid < 0: semantic = self.__view(top, (cid,))[0]
		else:
			args = self.__view(top, view)
			semantic = self._combine(cid, args)
		while length > 0:
			length -= 1
			top = top[self.NODE_PRIOR]
		return self._table.get_goto(top[self.NODE_STATE], nonterminal_id), top, semantic
	
	def __view(self, top, view):
		"""
		Recall that each element of view is a negative offset from the end of a notional
		linked-list-style stack, so in particular -1 is top-of-stack, and also these are
		presently constrained to appear in increasing order (starting negative and
		growing closer to zero).
		"""
		result = []
		depth = -1
		for seeking in reversed(view):
			while depth > seeking:
				depth -= 1
				top = top[self.NODE_PRIOR]
			result.append(top[self.NODE_SEMANTIC])
		result.reverse()
		return result



"""
By now we've seen two ways each to do both recognition and semantic recovery:

In module parsing.shift_reduce, the trial_parse(...) and parse(...) functions
show the linear-time approach appropriate to unambiguous parse tables -- i.e.
those which either have no CF-grammar inadequacies, or for which all such are
suitably resolved with operator-precedence/associativity declarations.

In module parsing.automata, the trial_parse(...) method of class HFA shows
a simple recognizer, whereas parsing.general.brute_force shows a simple
semantic recovery method, both suitable principally for getting our feet wet
with the issues involved in dealing with ambiguous parse tables.

It's time to do better.

In 1988, Tomita showed the world how to use a "graph-structured stack" to
avoid the exponential behaviour associated with the brute-force approach.
Tomita's original algorithm did not work properly with epsilon rules. Getting
them absolutely right is more than a bit tricky. Nozohoor-Farshi contributed
a method. I'm not 100% sure the method here is identical, but it does work.

This module contains that recognizer.

There are two things to bear in mind:

First,

Secondly, timings showed the brute-force approach to be slightly faster than
this for inputs with only mild levels of ambiguity, providing the grammars
don't have hidden left recursion (which sends the brute-force method into
an infinite loop).
"""

from typing import Dict
from ...support import interfaces


class GNode:
	# Nodes in the main portion of a graph-structured stack are simple enough.
	# You need to know what state number they are, and what their predecessor linkages are.
	# Any semantic values are associated with the linkages, not with the states themselves.
	# For this application I'm not second-guessing Python and just using dictionaries.
	state_id: int
	arcs: Dict["GNode", object]
	
	def __init__(self, state_id:int, arcs:Dict["GNode", object]):
		self.state_id, self.arcs = state_id, arcs
	
	def all_paths(self, depth: int):
		"""
		A previous version of this subroutine had potentially exponential
		behavior. It was just a naive depth-first traversal. A GSS can
		branch and merge, so a level-by-level approach seems better.
		"""
		frontier = {self}
		for _ in range(depth):
			frontier = set().union(*(n.arcs for n in frontier))
		return frontier
	
	def __repr__(self): return "<%d / %s>" % (self.state_id, ",".join(str(e.state_id) for e in self.arcs))

def gss_trial_parse(table: interfaces.ParseTable, sentence, *, language=None):
	def act_on(node: GNode, step: int):
		if step == 0: return
		elif step < 0: primary_reduction(node, -1-step)
		elif step < nr_states: shifts.append((node, step))
		else: split_stacks(node, table.get_split(step - nr_states))
	def split_stacks(node:GNode, steps):
		for s in steps: act_on(node, s)
	
	def primary_reduction(reach:GNode, rule_id:int):
		ners, fwd = books[reach]
		nonterminal_id, length, cid, view = table.get_rule(rule_id)
		if length: ners.append(rule_id)
		for prior_node in cook_paths(reach, length):
			perform_goto(prior_node, nonterminal_id)
	
	def secondary_reduction(chop:int, via:GNode, rule_id:int):
		nonterminal_id, length, cid, view = table.get_rule(rule_id)
		if length >= chop:
			for prior_node in cook_paths(via, length-chop):
				# Presumably capturing the parse tree would involve the REACH node somehow.
				perform_goto(prior_node, nonterminal_id)
	
	def perform_goto(prior_node:GNode, nonterminal_id:int):
		goto_id = table.get_goto(prior_node.state_id, nonterminal_id)
		if goto_id in arena:
			goto_node = arena[goto_id]
			if prior_node in goto_node.arcs: print('pun')  # Manage puns here.
			else:
				goto_node.arcs[prior_node] = None
				if goto_node in books: secondary.append((goto_node, prior_node))
		else:
			arena[goto_id] = GNode(goto_id, {prior_node: None})
			frontier.append(goto_id)
	
	def cook_paths(origin:GNode, depth:int):
		"""
		Why not just blindly take all paths N steps back from the origin?
		Because I want to capture certain data along the way: this data is
		only rarely useful, and only if the grammar has certain features.
		"""
		if depth == 0: yield origin
		else:
			for prior in list(origin.arcs):
				if prior in books:
					fwd = books[prior][1]
					if origin not in fwd: fwd.append(origin)
					yield from cook_paths(prior, depth-1)
				else: yield from prior.all_paths(depth-1)
	
	def drain_queues():
		shifts.clear()
		books.clear()
		frontier.extend(arena.keys())
		while frontier:
			drain_frontier()
			drain_secondary()
	
	def drain_frontier():
		while frontier:
			state_id = frontier.pop()
			node = arena[state_id]
			books[node] = ([], [])
			act_on(node, table.get_action(state_id, terminal_id))
	
	def drain_secondary():
		while secondary:
			goto_node, via = secondary.pop()
			rq, chop, ahead = [goto_node], 1, set()
			while rq:
				for node in rq:
					ners, fwd = books[node]
					for rule_id in ners: secondary_reduction(chop, via, rule_id)
					ahead.update(fwd)
				rq, chop, ahead = ahead, chop+1, set()
				
		
	
	def apply_shifts(semantic):
		arena.clear()
		for node, step in shifts:
			assert isinstance(node, GNode), type(node)
			assert isinstance(step, int), type(step)
			if step not in arena: arena[step] = GNode(step, {})
			arena[step].arcs[node] = semantic
	
	nr_states = table.get_split_offset()
	q0 = table.get_initial(language)
	arena = {q0: GNode(q0, {})}
	shifts, frontier, books, secondary = [], [], {}, []
	for symbol in sentence:
		terminal_id = table.get_translation(symbol)
		drain_queues()
		apply_shifts(None)
		if not shifts: raise interfaces.GeneralizedParseError("Parser died midway at something ungrammatical.")
	# Now deal with the end-of-input:
	terminal_id = 0
	drain_queues()
	if not shifts: raise interfaces.GeneralizedParseError("Parser recognized a viable prefix, but not a complete sentence.")
	return True





""" Small is beautiful. These algorithms need no introduction. """

from collections.abc import Sequence
import operator

def allocate(a_list:list, item):
	""" Too frequent an idiom not to abbreviate. """
	idx = len(a_list)
	a_list.append(item)
	return idx

def transitive_closure(roots, successors) -> set:
	""" Transitive closure is a simple unordered graph exploration. """
	black = set()
	grey = set(roots)
	while grey:
		k = grey.pop()
		if k not in black:
			black.add(k)
			them = successors(k)
			if them is not None: grey.update(them)
	return black

class BreadthFirstTraversal:
	""" This object also accumulates and exposes data about the traversal path. """
	def __init__(self):
		self.current, self.traversal, self.catalog, self.earliest_predecessor, self.breadcrumbs = None, [], {}, [], []
	def execute(self, visit):
		""" visit(key, lookup) should call self.lookup(successor_key, breadcrumb), which returns an integer. """
		for self.current, key in enumerate(self.traversal): visit(key)
		self.current = None
	def lookup(self, key, *, breadcrumb=None) -> int:
		if key not in self.catalog:
			self.catalog[key] = allocate(self.traversal, key)
			self.earliest_predecessor.append(self.current)
			self.breadcrumbs.append(breadcrumb)
		return self.catalog[key]
	def depth_list(self) -> list:
		result = []
		for p in self.earliest_predecessor: result.append(0 if p is None else 1+result[p])
		return result
	def shortest_path_to(self, index:int) -> list:
		""" Return a minimal list of states traversed, from a root to the given node index, in normal order. """
		path = []
		while index is not None:
			path.append(index)
			index = self.earliest_predecessor[index]
		path.reverse()
		return path

class EquivalenceClassifier:
	def __init__(self):
		self.catalog = {}
		self.exemplars = []
	def classify(self, key):
		if key not in self.catalog:
			self.catalog[key] = allocate(self.exemplars, key)
		return self.catalog[key]

def hamming_distance(a, b):
	""" Compute the number of places in two (equal-length) sequences with unequal corresponding elements. """
	assert len(a) == len(b)
	return sum(map(operator.ne, a, b))

def grade(seq:Sequence, *, descending=False) -> list:
	""" returns result such that [seq[i] for i in grade(seq)] == sorted(seq). Cribbed from APL's grade_up/down. """
	return sorted(range(len(seq)), key=seq.__getitem__, reverse=descending)

def everted(permutation:Sequence) -> list:
	""" Creates and returns an inverse permutation: like it's been turned inside-out """
	result = [None] * len(permutation)
	for i, x in enumerate(permutation): result[x] = i
	return result

def collation(seq:Sequence, *, descending=False):
	""" The inverse of `grade`: Tells you how a sorted sequence was permuted to get here. """
	return everted(grade(seq, descending=descending))

def strongly_connected_components_by_tarjan(graph):
	"""
	See https://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
	Returns a list of strongly-connected components in reverse topological order.
	Each component is a list of member node numbers. Deviating slightly from the wikipedia
	presentation, the low-link is local to the recursive call, eliminating one confusion.
	
	It's expected that graph[q] is the list of arcs from (or perhaps to) node q.
	The linear-time bound assumes all nodes are numbered from 0..last. An isomorphism
	for hashable node keys is provided below.
	"""
	def unvisited(q): return index[q] is None
	def connect(q) -> int:
		low_link = index[q] = allocate(stack, q)
		on_stack[q] = True
		for r in graph[q]:
			if unvisited(r): low_link = min(low_link, connect(r))
			elif on_stack[r]: low_link = min(low_link, index[r])
		if low_link == index[q]: # i.e. if node q is the root of an SCC:
			component = stack[low_link:]
			del stack[low_link:]
			for r in component: on_stack[r] = False
			output.append(component)
		return low_link
	def main():
		for q in range(size):
			if unvisited(q): connect(q)
	size = len(graph)
	index = [None] * size
	on_stack = [False] * size
	stack = []
	output = []
	main()
	return output

def strongly_connected_components_hashable(graph:dict):
	""" This adapts Tarjan's SCC algorithm for more kinds of graph node labels than strictly integers. """
	table = list(graph.keys())
	index = {key:i for i,key in enumerate(table)}
	prime = [
		[index[arc] for arc in node if arc in index]
		for node in graph.values()
	]
	return [[table[q] for q in component] for component in strongly_connected_components_by_tarjan(prime)]


class Visitor:
	"""
	Visitor-pattern in Python, with fall-back to superclasses along the MRO.

	Actual visitation-algorithms will inherit from Visitor and then each
	`visit_Foo` method must call `self.visit(host.bar)` as appropriate. This
	is so that your visitation-algorithm is in control of which bits of an
	object-graph that it actually visits, and in what order.
	
	Does this really belong here? Perhaps it lacks the gravitas of classical
	algorithms, but it's certainly generic enough and useful in structure-
	directed applications.
	"""
	
	def visit(self, host, *args, **kwargs):
		method_name = 'visit_' + host.__class__.__name__
		try: method = getattr(self, method_name)
		except AttributeError:
			# Searching the MRO incurs whatever cost there is to set up an iterator.
			# NB: Multiple-inheritance with NamedTuple seems to confuse the __mro__.
			for cls in host.__class__.__mro__:
				fallback = 'visit_' + cls.__name__
				if hasattr(self, fallback):
					method = getattr(self, fallback)
					break
			else: raise
		return method(host, *args, **kwargs)




""" An AST class-hierarchy for regular expressions and mechanisms for translating them. """

from typing import List
from ..support import foundation
from . import finite, charset

class PatternError(Exception):
	"""
	Raised if something wrong is found with the semantics of a regular expression.
	"""
class BadReferenceError(PatternError):
	"""
	Raised if a pattern refers to a subexpression name not defined in the context of the analysis.
	args are the offending reference and the span within the pattern (offset/length pair) where it appears.
	"""
class CircularDefinitionError(PatternError):
	"""
	Raised if a named pattern somehow refers back to itself.
	args are the offending reference and the span within the pattern (offset/length pair) where it appears.
	"""
class NonClassError(PatternError):
	"""
	Raised if a character class definition refers to a named expression which is not itself a character class.
	args are the offending reference and the span within the pattern (offset/length pair) where it appears.
	"""
class TrailingContextError(PatternError):
	"""
	Raised if pattern has both variable-sized stem and variable-sized trailing context.
	This is not supported at this time.
	"""


class Regular: pass


class CharClass(Regular): pass

class CharSpecial(CharClass):
	""" Useful for pre-built classes, I suppose... """
	def __init__(self, cls:List[int]):
		self.cls = cls

class Letter(CharClass):
	def __init__(self, codepoint:int):
		self.codepoint = codepoint

class CharRange(CharClass):
	def __init__(self, first:int, last:int):
		self.first, self.last = first, last

class CharUnion(CharClass):
	def __init__(self, a:CharClass, b:CharClass): self.a, self.b = a,b

class CharIntersection(CharClass):
	def __init__(self, a:CharClass, b:CharClass): self.a, self.b = a,b

class CharComplement(CharClass):
	def __init__(self, inverse:CharClass): self.inverse = inverse

class Binary(Regular):
	def __init__(self, a:Regular, b:Regular): self.a, self.b = a,b

class Alternation(Binary): pass
class Sequence(Binary): pass
class Inflection(Regular):
	def __init__(self, sub:Regular): self.sub = sub

class Star(Inflection): pass
class Hook(Inflection): pass
class Plus(Inflection): pass
class Counted(Regular):
	""" Just too handy not to provide. """
	def __init__(self, sub: Regular, m:int, n):
		assert isinstance(sub, Regular)
		assert isinstance(m, int)
		assert isinstance(n, int) or n is None
		self.sub, self.m, self.n = sub, m, n

class NamedSubexpression(Regular):
	def __init__(self, name, trace):
		self.name, self.trace = name, trace


class Encoder(foundation.Visitor):
	"""
	This visitor represents the strategy to encode a regular expression as
	a non-deterministic finite state automaton.
	"""
	def __init__(self, nfa:finite.NFA, rank:int, names:dict):
		self.__nfa = nfa
		self.__rank = rank
		self.__names = names
		self.__ce = ClassEncoder(names)
		self.__inside = set()  # For detecting circular definitions.
	def __new_node(self): return self.__nfa.new_node(self.__rank)
	def __eps(self, src:int, dst:int, ): self.__nfa.link_epsilon(src, dst)
	def visit_CharClass(self, cc:CharClass, src:int, dst:int, ):
		self.__nfa.link(src, dst, self.__ce.visit(cc))
	def visit_Alternation(self, alt:Alternation, src:int, dst:int, ):
		self.visit(alt.a, src, dst)
		self.visit(alt.b, src, dst)
	def visit_Sequence(self, seq:Sequence, src:int, dst:int, ):
		midpoint = self.__new_node()
		self.visit(seq.a, src, midpoint)
		self.visit(seq.b, midpoint, dst)
	def visit_Star(self, star:Star, src:int, dst:int, ):
		loop = self.__new_node()
		self.visit(star.sub, loop, loop)
		self.__eps(src, loop)
		self.__eps(loop, dst)
	def visit_Hook(self, hook:Hook, src:int, dst:int, ):
		self.visit(hook.sub, src, dst)
		self.__eps(src, dst)
	def visit_Plus(self, plus:Plus, src:int, dst:int, ):
		before, after = self.__new_node(), self.__new_node()
		self.visit(plus.sub, before, after)
		self.__eps(src, before)
		self.__eps(after, before)
		self.__eps(after, dst)
	def visit_Counted(self, ct:Counted, src:int, dst:int, ):
		p1 = self.__new_node()
		self.__eps(src, p1)
		for _ in range(ct.m):
			p2 = self.__new_node()
			self.visit(ct.sub, p1, p2)
			p1 = p2
		self.__eps(p1, dst)
		if ct.n is None:
			self.visit(ct.sub, p1, p1)
		else:
			for _ in range(ct.n - ct.m):
				p2 = self.__new_node()
				self.visit(ct.sub, p1, p2)
				self.__eps(p2, dst)
				p1 = p2
	def visit_NamedSubexpression(self, ns:NamedSubexpression, src:int, dst:int, ):
		try: item = self.__names[ns.name]
		except KeyError: raise BadReferenceError(ns.name, ns.trace)
		if ns.name in self.__inside: raise CircularDefinitionError(ns.name, ns.trace)
		self.__inside.add(ns.name)
		result = self.visit(item, src, dst)
		self.__inside.remove(ns.name)
		return result


class Sizer(foundation.Visitor):
	"""
	Finds the fixed length of a Regular, if it is defined.
	"""
	
	def __init__(self, names: dict):
		self.__names = names
	
	def visit_CharClass(self, _): return 1
	def visit_Alternation(self, alt:Alternation):
		a, b = self.visit(alt.a), self.visit(alt.b)
		return a if a == b else None
	def visit_Sequence(self, seq:Sequence):
		a, b = self.visit(seq.a), self.visit(seq.b)
		if None not in (a,b): return a + b
	def visit_Star(self, star: Star): return None
	def visit_Hook(self, hook: Hook): return None
	def visit_Plus(self, plus: Plus): return None
	def visit_Counted(self, ct: Counted):
		if ct.m == ct.n:
			x = self.visit(ct.sub)
			if x is not None: return x * ct.m
	def visit_NamedSubexpression(self, ns:NamedSubexpression):
		try: item = self.__names[ns.name]
		except KeyError: raise BadReferenceError(ns.name, ns.trace)
		return self.visit(item)


class ClassEncoder(foundation.Visitor):
	"""
	Builds a character class in the format used by the finite.NFA class.
	
	One might think this could all be done inside class Encoder, but the
	treatment of name references differs.
	"""
	def __init__(self, names:dict):
		self.__names = names
	
	def visit_Letter(self, ltr:Letter): return charset.singleton(ltr.codepoint)
	def visit_CharSpecial(self, cs:CharSpecial): return cs.cls
	def visit_CharRange(self, cr:CharRange): return charset.range_class(cr.first, cr.last)
	def visit_CharUnion(self, cu:CharUnion): return charset.union(self.visit(cu.a), self.visit(cu.b))
	def visit_CharIntersection(self, ci:CharIntersection): return charset.intersect(self.visit(ci.a), self.visit(ci.b))
	def visit_CharComplement(self, cc:CharComplement): return charset.complement(self.visit(cc.inverse))
	def visit_NamedSubexpression(self, ns:NamedSubexpression):
		try: expr = self.__names[ns.name]
		except KeyError: raise BadReferenceError(ns.name, ns.trace)
		if isinstance(expr, (CharClass, NamedSubexpression)): return self.visit(expr)
		else: raise NonClassError(ns.name, ns.trace)
	


from . import interfaces, runtime
from ..scanning import miniscan
from ..parsing import miniparse

class TypicalMiniApp(runtime.AbstractTypical):
	"""
	This class specializes for applications founded on mini-scan and mini-parse,
	to provide reasonable default error handling behavior.
	"""
	
	def __init__(self, *, scan:miniscan.Definition, parse:miniparse.MiniParse):
		hfa, combine = parse.get_hfa_and_combine()
		super().__init__(
			dfa=scan.get_dfa(),
			scan_rules=scan,
			hfa=hfa,
			combine=combine,
		)




""" Hook regular expression patterns up to method calls on a scanner object. """
from boozetools.support.interfaces import Scanner

from ..support import interfaces
from ..parsing import miniparse
from . import finite, regular, charset, recognition

PRELOAD = {'ASCII': {k: regular.CharSpecial(cls) for k, cls in charset.mode_ascii.items()}}


class Definition(interfaces.ScanRules):
	def __init__(self, *, minimize=True, mode='ASCII'):
		self.__actions = []
		self.__trails = []
		self.__subexpressions = PRELOAD[mode].copy()
		self.__dfa: finite.DFA = None
		self.__nfa = finite.NFA()
		self.__minimize = minimize
		self.__awaiting_action = False
	
	def get_trailing_context(self, rule_id: int): return self.__trails[rule_id]
	
	def get_dfa(self) -> interfaces.FiniteAutomaton:
		if self.__dfa is None:
			if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the final pattern!')
			self.__dfa = self.__nfa.subset_construction()
			if self.__minimize: self.__dfa = self.__dfa.minimize_states().minimize_alphabet()
		return self.__dfa
	
	def scan(self, text, *, start=None, on_error:interfaces.ScanErrorListener = None):
		if on_error is None: on_error = interfaces.ScanErrorListener()
		scanner = recognition.IterableScanner(text=text, automaton=self.get_dfa(), rules=self, start=start, on_error=on_error)
		return scanner
		
	def install_subexpression(self, name:str, expression: regular.Regular):
		assert isinstance(name, str) and name not in self.__subexpressions and len(name) > 1
		assert isinstance(expression, regular.Regular)
		self.__subexpressions[name] = expression
	
	def install_rule(self, *, action:callable, expression: regular.Regular, bol=(True, True), condition: (str, list, tuple)=None, trail:int=None, rank:int=0) -> int:
		rule_id = len(self.__actions)
		self.__actions.append(action)
		self.__trails.append(trail)
		src = self.__nfa.new_node(rank)
		dst = self.__nfa.new_node(rank)
		if condition is None or isinstance(condition, str): condition = [condition]
		else: assert isinstance(condition, (tuple, list)), type(condition)
		for C in condition:
			for q, b in zip(self.__nfa.condition(C), bol):
				if b: self.__nfa.link_epsilon(q,src)
		self.__nfa.final[dst] = rule_id
		regular.Encoder(self.__nfa, rank, self.__subexpressions).visit(expression, src, dst)
		return rule_id
	
	def let(self, name:str, pattern:str):
		self.install_subexpression(name, rex.parse(META.scan(pattern), language='Regular'))
	
	def token(self, kind:str, pattern:str, *, rank=0, condition=None):
		""" This says every member of the pattern has token kind=kind and semantic=matched text. """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:Scanner): yy.token(kind, yy.matched_text())
	
	def token_map(self, kind:str, pattern:str, fn:callable, *, rank=0, condition=None):
		""" Every member of the pattern has token kind=kind and semantic=fn(matched text). """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:Scanner): yy.token(kind, fn(yy.matched_text()))

	def ignore(self, pattern:str, *, rank=0, condition=None):
		""" Tell Scanner to ignore what matches the pattern. """
		@self.on(pattern, rank=rank, condition=condition)
		def action(yy:Scanner): pass

	def on(self, pattern:str, *, condition=None, rank=0):
		"""
		For instance:
		scanner_definition.on('r#[^\n]+')(None) # Ignore comments
		@scanner_definition.on(r'[A-Za-z_]+')
		def word(yy): yy.token('word', scanner.matched_text())
		"""
		if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the previous pattern!')
		self.__awaiting_action = True
		bol, expression, trail = analyze_pattern(pattern, self.__subexpressions)
		def decorator(fn):
			assert self.__awaiting_action
			self.__awaiting_action = False
			assert callable(fn)
			self.install_rule(action=fn, expression=expression, bol=bol, condition=condition, trail=trail, rank=rank)
			return fn
		return decorator
	def condition(self, *condition): return ConditionContext(self, *condition)
	
	def invoke(self, yy: interfaces.Scanner, rule_id:int):
		action = self.__actions[rule_id]
		if callable(action): action(yy)
		else: assert action is None
	


def analyze_pattern(pattern:str, env):
	scanner = META.scan(pattern)
	bol, expression, trailing_context = rex.parse(scanner)
	assert isinstance(expression, regular.Regular)
	if not trailing_context: trail = None
	else:
		assert isinstance(trailing_context, regular.Regular), trailing_context
		sizer = regular.Sizer(env)
		stem, trail = sizer.visit(expression), sizer.visit(trailing_context)
		expression = regular.Sequence(expression, trailing_context)
		if trail: trail = -trail
		elif stem: trail = stem
		else: raise regular.TrailingContextError('Variable stem and variable trailing context in the same pattern are not presently supported.')
	return bol, expression, trail

class ConditionContext:
	""" I'd like to be able to use Python's context manager protocol to simplify writing definitions of scan conditions. """
	def __init__(self, definition:Definition, *condition):
		self.__definition = definition
		self.__condition = condition
	def __enter__(self): return self
	def __exit__(self, exc_type, exc_val, exc_tb): pass
	def on(self, pattern, *, rank=0): return self.__definition.on(pattern, rank=rank, condition=self.__condition)
	def install_rule(self, **kwargs): return self.__definition.install_rule(condition=self.__condition, **kwargs)
	def token(self, kind:str, pattern:str, *, rank=0): self.__definition.token(kind, pattern, rank=rank, condition=self.__condition)
	def token_map(self, kind:str, pattern:str, fn:callable, *, rank=0): self.__definition.token_map(kind, pattern, fn, rank=rank, condition=self.__condition)
	def ignore(self, pattern:str, *, rank=0): self.__definition.ignore(pattern, rank=rank, condition=self.__condition)

#########################
# A pattern parser is easy to build using the miniparse module:

rex = miniparse.MiniParse('Pattern', 'Regular')
rex.void_symbols.update('^ ^^ / | ? * + { } , ( ) [ - ] &&'.split())
rex.rule('Pattern', 'BOL Regular Trail')()
rex.rule('Pattern', 'BOL $')(lambda bol, eof:(bol, eof, None)) # To specify end-of-file rules without introducing yet another metacharacter
rex.rule('BOL', '')(lambda : (True, True))
rex.rule('BOL', '^')(lambda : (False, True))
rex.rule('BOL', '^^')(lambda : (True, False))
rex.rule('Trail', '')()
rex.rule('Trail', '$')()
rex.rule('Trail', '/ Regular')()
rex.rule('Trail', '/ Regular $')(regular.Sequence)
rex.rule('Regular', 'Sequence')()
rex.rule('Regular', 'Regular | Sequence')(regular.Alternation)
rex.rule('Sequence', 'Term')()
rex.rule('Sequence', 'Sequence Term')(regular.Sequence)
rex.rule('Term', 'Atom')()
rex.rule('Term', 'Atom ?')(regular.Hook)
rex.rule('Term', 'Atom *')(regular.Star)
rex.rule('Term', 'Atom +')(regular.Plus)
rex.rule('Term', 'Atom { number }')(lambda a, n: regular.Counted(a, n, n))
rex.rule('Term', 'Atom { number , }')(lambda a, n: regular.Counted(a, n, None))
rex.rule('Term', 'Atom { , number }')(lambda a, n: regular.Counted(a, 0, n))
rex.rule('Term', 'Atom { number , number }')(lambda a, m, n: regular.Counted(a, m, n))
rex.rule('Atom', '( Regular )')()
rex.rule('Atom', 'c')(regular.Letter)
rex.rule('Atom', 'reference')()
rex.rule('Atom', '[ Class ]')()
rex.rule('Class', 'Conjunct')()
rex.rule('Class', 'Class && Conjunct')(regular.CharIntersection)
rex.rule('Conjunct', 'Members')()
rex.rule('Conjunct', '^ Members')(regular.CharComplement)
rex.rule('Members', 'Item')()
rex.rule('Members', 'Members Item')(regular.CharUnion)
rex.rule('Item', 'c')(regular.Letter)
rex.rule('Item', 'c - c')(regular.CharRange)
rex.rule('Item', 'short')()
rex.rule('Item', 'reference')()

META = Definition()
def _BEGIN_():
	""" This is the bootstrapping routine: it builds the preload and the meta-scanner. """
	def seq(head, *tail):
		for t in tail: head = regular.Sequence(head, t)
		return head
	def txt(s):return seq(*(regular.CharSpecial(charset.singleton(ord(_))) for _ in s))
	
	def _metatoken(yy): yy.token(yy.matched_text(), None)
	def _and_then(condition):
		def fn(yy):
			_metatoken(yy)
			yy.enter(condition)
		return fn
	def _instead(condition):
		def fn(yy):
			yy.less(0)
			yy.enter(condition)
		return fn
	def _bracket_reference(yy:interfaces.Scanner):
		name = yy.matched_text()[1:-1]
		node = regular.NamedSubexpression(name, yy.current_span())
		yy.token('reference', node)
	def _shorthand_reference(yy:interfaces.Scanner):
		yy.token('reference', regular.NamedSubexpression(yy.matched_text()[1], yy.current_span()))
	def _dot_reference(yy:interfaces.Scanner):
		yy.token('reference', regular.NamedSubexpression('DOT', yy.current_span()))
	def _hex_escape(yy): yy.token('c', int(yy.matched_text()[2:], 16))
	def _control(yy): yy.token('c', 31 & ord(yy.matched_text()[2:]))
	def _arbitrary_character(yy): yy.token('c', ord(yy.matched_text()))
	def _class_initial_close_bracket(yy):
		yy.enter('in_class')
		_arbitrary_character(yy)
	def _class_final_dash(yy):
		yy.token('c', ord('-'))
		yy.token(']', None)
		yy.enter(None)
	def _arbitrary_escape(yy): yy.token('c', ord(yy.matched_text()[1:]))
	def _number(yy): yy.token('number', int(yy.matched_text()))
	def _dollar(charclass):
		def fn(yy:Scanner): yy.token('$', charclass)
		return fn
	def _meta_caret(yy):
		yy.token(yy.matched_text()[:-1], None)
		yy.token('^')
		yy.enter('start_class')
	
	def common_rules(ctx, *rules):
		for lit, act in rules: ctx.install_rule(expression=txt(lit), action=act)
	
	def ref(x): return PRELOAD['ASCII'][x]
	
	dot = ref('DOT')
	
	eof_charclass = regular.CharSpecial(charset.EOF)
	dollar_charclass = regular.CharSpecial(charset.union(charset.EOF, PRELOAD['ASCII']['vertical'].cls))
	
	for t in '^', '^^': META.install_rule(expression=txt(t), action=_metatoken, bol=(False, True))
	for t,cc in ('$', dollar_charclass), ('<<EOF>>', eof_charclass):
		META.install_rule(expression=seq(txt(t), eof_charclass), trail=-1, action=_dollar(cc))
	common_rules(META,
		('.', _dot_reference),
		('{', _and_then('brace')),
		('[', _and_then('start_class')),
		('[^', _meta_caret),
		*((c, _metatoken) for c in '(|)?*+/'),
	)
	common_rules(META.condition('in_class'),
		(']', _and_then(None)),
		('&&', _metatoken),
		('&&^', _meta_caret),
		('-', _metatoken),
		('-]', _class_final_dash),
	)
	with META.condition('start_class') as start_class:
		common_rules(start_class, (']', _class_initial_close_bracket), ('-', _arbitrary_character),)
		start_class.install_rule(expression=ref('ANY'), action = _instead('in_class'))
	with META.condition(None, 'in_class') as anywhere:
		anywhere.install_rule(expression=seq(txt('{'), ref('alpha'), regular.Plus(ref('word')), txt('}'), ), action=_bracket_reference)
		whack = txt('\\') # NB: Python doesn't let you end a raw-string with a backslash.
		for c, n in [('x', 2), ('u', 4), ('U', 8)]:
			anywhere.install_rule(expression=seq(whack, txt(c), regular.Counted(ref('xdigit'), n, n)), action=_hex_escape)
		anywhere.install_rule(expression=seq(whack, txt('c'), regular.CharSpecial(charset.range_class(64, 127))), action=_control)
		anywhere.install_rule(expression=seq(whack, ref('alnum')), action=_shorthand_reference)
		anywhere.install_rule(expression=seq(whack, dot), action=_arbitrary_escape)
		anywhere.install_rule(expression=dot, action=_arbitrary_character)
	with META.condition('brace') as brace:
		brace.install_rule(expression=regular.Plus(ref('digit')), action=_number)
		common_rules(brace, (',', _metatoken), ('}', _and_then(None)),)
	
	PRELOAD['ASCII']['R'] = rex.parse(META.scan(r'\r?\n|\r'), language='Regular')
_BEGIN_()
